-- Datetime Calculations

-- Sep 1, 1996
-- written by Junko C. Miura, Rapid Deployment Software
-- Contributed to the Public Domain

-- Aug 5, 2010
-- converted by Jean-Marc Duro to a reusable unit

-- Nov 18, 2015
-- added strDateToDatetime, secondsToStrTime, excelTimeToDatetime, datetimeToExcelTime

include myDebug.e
include myConv.e
include myMath.e
include mySeq.e as seq

global constant
  YEAR = 1,
  MONTH = 2,
  DAY = 3,
  HOUR = 4,
  MINUTE = 5,
  SECOND = 6

-- Abbreviations of month names

global constant month_abbrs = {
	"Jan", "Feb", "Mar", "Apr", "May", "Jun",
	"Jul", "Aug", "Sep", "Oct", "Nov", "Dec"
}

------------------------------------------------------------------------------

function INT(atom x)
  if x >= 0 then
    return floor(x)
  else
    return -1 * floor(-x)
  end if
end function

------------------------------------------------------------------------------

global function getFactor(integer yyyy, integer mm, integer dd)
  integer factor
     
  if yyyy <= 1582 then
	warnError("year must be >1582", 0)
  end if
  if mm < 1 or mm > 12 then
	warnError("month must be >0 and <=12", 0)
  end if
  if mm < 1 or mm > 31 then
	warnError("day must be >0 and <=31", 0)
  end if
  if mm > 2 then
    factor = 365 * yyyy + dd + 31 * (mm - 1) - INT(0.4 * mm + 2.3)
		 + INT(yyyy / 4) - INT(0.75 * (INT(yyyy / 100) + 1))
  else
	factor = 365 * yyyy + dd + 31 * (mm - 1) + INT((yyyy - 1) / 4)
		 - INT(0.75 * (INT(((yyyy - 1) / 100) + 1)))
  end if
  return factor
end function

------------------------------------------------------------------------------

global function dayOfWeek(integer yyyy, integer mm, integer dd)
  integer factor, dayOfWeek
    
  factor = getFactor(yyyy, mm, dd)
  dayOfWeek = factor + (INT(-factor / 7) * 7)
  if dayOfWeek > 6 then              -- start of DEBUG
	warnError(sprintf("WARNING-- dayOfWeek : %d\n", dayOfWeek), 0)
	dayOfWeek = remainder(dayOfWeek, 7) 
  end if                             -- end of DEBUG
  return dayOfWeek + 1
end function

------------------------------------------------------------------------------

global function getDaysBetwDates(sequence date1, sequence date2)
  integer factor1, factor2, numDays
    
  factor1 = getFactor(date1[1],date1[2],date1[3])
  factor2 = getFactor(date2[1],date2[2],date2[3])
  if factor2 > factor1 then
    numDays = factor2 - factor1
  else
    numDays = factor1 - factor2
  end if
  return numDays
end function

------------------------------------------------------------------------------

global function strDateToDatetime(sequence oa)
  sequence s
  integer y, m, d
  
  s = seq:split(oa, " ", 0)
  m = find(s[2], month_abbrs)
  if m = 0 then
    return "Invalid Month"
  end if
  y = val(s[3])
  d = val(s[1])
  s = seq:split(s[4], ":", 0)
  if length(s) = 3 then
    return {y, m, d, val(s[1]), val(s[2]), val(s[3])}
  else
    return {y, m, d, val(s[1]), val(s[2]), 0}
  end if
end function

------------------------------------------------------------------------------

global function secondsToStrTime(atom t)
  integer j, h, m, s

  j = floor(t/86400)
  t -= j*86400
  h = floor(t/3600)
  t -= h*3600
  m = floor(t/60)
  t -= m*60
  s = floor(t)
  return sprintf("%02dj %02dh %02dmn %02ds", {j, h, m, s})
end function

------------------------------------------------------------------------------

global function excelTimeToDatetime(atom a)
  atom t
  integer ly
  sequence dt
  
  -- return date
  dt = {1900,1,1,0,0,0}
  a = a - 1  -- 1/1/1900 begins as #1
  while a > 0 do
    if remainder(dt[YEAR],4)=0 then
      -- and (remainder(dt[YEAR],100)>0 or remainder(dt[YEAR],400)=0) Microsoft Excel includes years divisible
      -- by 100 but not by 400 as leap years.
      ly = 1
    else
      ly = 0
    end if
    if a >= 366 and ly = 1 then
      dt[YEAR] = dt[YEAR] + 1
      a = a - 366
    elsif a >= 365 and ly = 0 then
      dt[YEAR] = dt[YEAR] + 1
      a = a - 365
    elsif find(dt[MONTH],{1,3,5,7,8,10,12}) and a >= 31 then
      dt[MONTH] = dt[MONTH] + 1
      a = a - 31
    elsif find(dt[MONTH],{4,6,9,11}) and a >= 30 then
      dt[MONTH] = dt[MONTH] + 1
      a = a - 30
    elsif dt[MONTH]=2 and ly=1 and a >= 29 then
      dt[MONTH] = dt[MONTH] + 1
      a = a - 29
    elsif dt[MONTH]=2 and ly=0 and a >= 28 then
      dt[MONTH] = dt[MONTH] + 1
      a = a - 28
    else
      dt[DAY] = dt[DAY] + a
      a = 0
    end if
  end while
  t = frac(a) * 24
  dt[HOUR] = trunc(t)
  t = (t-dt[HOUR]) * 60
  dt[MINUTE] = trunc(t)
  t = (t-dt[MINUTE]) * 60
  dt[SECOND] = trunc(t)
  return dt
end function

------------------------------------------------------------------------------

global function datetimeToExcelTime(sequence dt)
  atom nb
  
  nb = getDaysBetwDates({1899,12,30}, {dt[YEAR],dt[MONTH],dt[DAY]})
  nb += (dt[SECOND]+dt[MINUTE]*60+dt[HOUR]*3600)/86400
  return nb 
end function

