-- mySeq.e
-- v1.0
-- 2010-11-13
-- sequence functions

-- function trim_head(sequence s)
--   removes space chars from the beginning of a string
-- function trim_tail(sequence s)
--   removes space chars from the end of a string
-- function trim(sequence s)
--   removes space chars from both ends of a string
-- function pad_tail(sequence s, atom len)
--   returns a string right-padded to length len
-- function pad_head(sequence s, atom len)
--   returns a string left-padded to length len
-- function right(sequence s, integer n)
--   returns the n characters at the right of string s

include myTypes.e
include myConv.e
include myDebug.e

------------------------------------------------------------------------------

global function trim_head(sequence s)
-- remove space chars from the beginning of a string
  integer l

  l = length(s)
  if l=0 then return "" end if
  for i=1 to l do
    if s[i] != 32 then
      return s[i..l]
    end if
  end for
  return ""
end function

------------------------------------------------------------------------------

global function trim_tail(sequence s)
-- remove space chars from the end of a string
  integer l

  l = length(s)
  if l=0 then return "" end if
  for i=l to 1 by -1 do
    if s[i] != 32 then
      return s[1..i]
    end if
  end for
  return ""
end function

------------------------------------------------------------------------------

global function trim(sequence s)
-- remove space chars from both ends of a string
  return trim_head(trim_tail(s))
end function

------------------------------------------------------------------------------

global function pad_tail(sequence s, atom len)
--Return a string right-padded to length len

  if length(s) >= len then
    return s
  else
    return s & repeat(' ', len-length(s))
  end if
end function

------------------------------------------------------------------------------

global function pad_head(sequence s, atom len)
--Return a string left-padded to length len

  if length(s) >= len then
    return s
  else
    return repeat(' ', len-length(s)) & s
  end if
end function

------------------------------------------------------------------------------

global function left(sequence s, integer n)
-- returns the n characters at the right of string s
  integer l

  l = length(s)
  if (l=0) or (n=0) then return "" end if
  if n >= l then return s end if
  return s[1..n]
end function

------------------------------------------------------------------------------

global function mid(sequence target, atom start, atom len)
  integer lg

  lg = length(target)
  if start > lg then
    return {}
  elsif len > lg then
    return target[start..lg]
  else
    return target[start..start+len-1]
  end if
end function

------------------------------------------------------------------------------

global function right(sequence s, integer n)
-- returns the n characters at the right of string s
  integer l

  l = length(s)
  if (l=0) or (n=0) then return "" end if
  if n >= l then return s end if
  return s[l-n+1..l]
end function

------------------------------------------------------------------------------

global function maxIndex(sequence lst)
-- returns index of the highest value from the list lst
-- if unsuccessful, returns 0
  atom maxi

  for i = 1 to length(lst) do
    if sequence(lst[i]) then
      return 0
    elsif i=1 then
      maxi = i
    elsif lst[i] > lst[maxi] then
      maxi = i
    end if
  end for
  return maxi
end function

------------------------------------------------------------------------------

global function minIndex(sequence lst)
-- returns index of the lowest value from the list lst
-- if unsuccessful, returns 0
  atom mini

  for i = 1 to length(lst) do
    if sequence(lst[i]) then
      return 0
    elsif i=1 then
      mini = i
    elsif lst[i] < lst[mini] then
      mini = i
    end if
  end for
  return mini
end function

------------------------------------------------------------------------------

global function quote(sequence str, object delims)
-- insert string between quotes
  if atom(delims) then
    return delims & str & delims
  elsif (length(delims) = 2) then
    return delims[1] & str & delims[2]
  end if
end function

------------------------------------------------------------------------------

global function dequote(sequence str, object delims)
-- remove specified quotes if existing
  if (length(str) > 2) then
    if atom(delims) then
      if (str[1]=delims) and (str[$]=delims) then
        return str[2..$-1]
      end if
    else
      if (length(delims) = 2) and (str[1]=delims[1]) and (str[$]=delims[2]) then
        return str[2..$-1]
      end if
    end if
  end if
  return str
end function

------------------------------------------------------------------------------

global function begins(object sub_text, sequence full_text)
  if length(full_text) = 0 then
    return 0
  end if
  
  if atom(sub_text) then
    if equal(sub_text, full_text[1]) then
      return 1
    else
      return 0
    end if
  end if
  
  if length(sub_text) > length(full_text) then
    return 0
  end if
  
  if equal(sub_text, full_text[1.. length(sub_text)]) then
    return 1
  else
    return 0
  end if
end function

------------------------------------------------------------------------------

global function ends(object sub_text, sequence full_text)
  if length(full_text) = 0 then
    return 0
  end if
  
  if atom(sub_text) then
    if equal(sub_text, full_text[$]) then
      return 1
    else
      return 0
    end if
  end if
  
  if length(sub_text) > length(full_text) then
    return 0
  end if
  
  if equal(sub_text, full_text[$ - length(sub_text) + 1 .. $]) then
    return 1
  else
    return 0
  end if
end function

------------------------------------------------------------------------------

-- split(sequence str, sequence sep, integer keep=0)
global function split(sequence str, sequence sep, integer keep)
-- splits str string in tokens according to separator sep
  integer slen
  sequence result, s
  integer tmpStr

  result = {}
  slen = length(str)
  if slen = 0 then return result end if
  
  s = {}
  tmpStr = 0      -- if 1, index is within double quotes

  for i = 1 to slen do
    if (str[i] = '"') and not find('"', sep) then
      s = s & str[i]
      if tmpStr = 0 then
        tmpStr = 1
      else
        tmpStr = 0
      end if
    else
      if tmpStr = 0 and find(str[i], sep) then
        result = append(result, trim(s))
        if keep then result = append(result, str[i]) end if
        s = {}
      else
        s = s & str[i]
      end if
    end if
  end for
  result = append(result, trim(s))
  return result
end function

------------------------------------------------------------------------------

global function join(sequence str, sequence sep)
  integer slen
  sequence result

  slen = length(str)
  if slen = 0 then return "" end if
  
  result = str[1]
  for i = 2 to slen do
    result &= sep & str[i]
  end for
  return result
end function

------------------------------------------------------------------------------

global function strtok(sequence str, sequence sep)
-- splits str string in tokens according to separator sep
-- converts segments to atoms when needed
  sequence s

  s = split(str, sep, 0)
  for i = 1 to length(s) do
    if is_number(s[i]) or is_hex(s[i]) then
      s[i] = val(s[i])
    elsif is_string(s[i]) then
      s[i] = dequote(s[i], {'"','"'})
    end if
  end for
  return s
end function

------------------------------------------------------------------------------

global function tokenize(sequence tableau, atom sep)
-- assembles tokens in a string with separator sep
  sequence s

  if atom(tableau[1]) then
    s = sprintf("%d", tableau[1])
  else
    s = tableau[1]
  end if
  for i = 2 to length(tableau) do
    if atom(tableau[1]) then
      s &= {sep}&sprintf("%d", tableau[i])
    else
      s &= {sep}&tableau[i]
    end if
  end for
--  puts(f_debug, s&"\n")
  return s
end function

------------------------------------------------------------------------------

global function insert(sequence target, object what, integer index)
-- inserts object what in sequence target at position index
-- returns modified sequence
  integer lg

  lg = length(target)
  if index <=1 then
    return what & target
  elsif index >= lg then
    return target & what
  else
    return target[1..index-1] & what & target[index..$]
  end if
end function

------------------------------------------------------------------------------

global function remove(sequence target, atom start, atom stop)
  integer lg

  lg = length(target)
  if (start<1) or (start>lg) then return target end if
  if start = 1 then
    return target[stop+1..lg]
  elsif start = lg then
    return target[1..lg-1]
  else
    return target[1..start-1] & target[stop+1..lg]
  end if
end function

------------------------------------------------------------------------------

global function replace(sequence target, sequence old, sequence new)
  integer index

  index = match(old,target)
  if index > 0 then
    target = remove(target,index,index+length(new)-1)
    target = insert(target,new,index)
  end if
  return target
end function

------------------------------------------------------------------------------

global function replace_all(sequence target, sequence old, sequence new)
-- replaces sub-sequence old by new in sequence target
  integer index, lg

  lg = length(old)
  index = -1
  while index != 0 do
    index = match(old, target)
    if index > 1 then
      target = target[1..index-1]&new&target[index+lg..$]
    elsif index = 1 then
      target = new&target[index+lg..$]
    end if
  end while
  return target
end function

------------------------------------------------------------------------------

global function find_all(sequence what, sequence target, integer from)
  integer index
  sequence result

  result = {}
  index = -1
  while index != 0 do
    index = find_from(what, target, from)
    if index then
      result = append(result, index)
      from = index+1
    end if
  end while
  return result
end function

------------------------------------------------------------------------------

global function match_all(sequence what, sequence target, integer from)
  integer index
  sequence result

  result = {}
  index = -1
  while index != 0 do
    index = match_from(what, target, from)
    if index then
      result = append(result, index)
      from = index+length(what)
    end if
  end while
  return result
end function

------------------------------------------------------------------------------

global function match_list(sequence what, sequence target, integer from)
  integer index

  for i = 1 to length(what) do
    index = match_from(what[i], target, from)
    if index then return index end if
  end for
  return 0
end function

------------------------------------------------------------------------------

global procedure hexDump(sequence s, integer output)
-- dumps a sequence in hex form to the specified output
  puts(output, "\n\nAddr  0  1  2  3  4  5  6  7  8  9  A  B  C  D  E  F\n")
  puts(output, "0000 ")
  for i = 1 to length(s) do
    puts(output, sprintf(" %02x", s[i]))
    if remainder(i, 16) = 0 then
      puts(output, sprintf("\n%04x ", floor(i/16)))
    end if
  end for
  puts(output, "\n\n")
end procedure

------------------------------------------------------------------------------

global function appendIfNew(sequence s, object o)
  if find(o, s) then
    return s
  else
    return append(s, o)
  end if
end function

-------------------------------------------------------------------------------

global function nbOccurences(sequence s, sequence item)
  integer nb, pos, lg

  nb = 0
  pos = 1
  lg = length(item)
  while pos > 0 do
    if match(item, s) then
      nb += 1
      s = remove(s, 1, pos+lg-1)
    end if
  end while
  return nb
end function
