-- Program then    TNCNX.PAS
-- Object then     Delphi component which implement the TCP/IP telnet protocol
--         including some options negociations.
--         RFC854, RFC885, RFC779, RFC1091
-- Author then     Franois PIETTE
-- EMail then    http then//users.swing.be/francois.piette  francois.piette@swing.be
--         http then//www.rtfm.be/fpiette       francois.piette@rtfm.be
--         francois.piette@pophost.eunet.be
-- Creation then   April, 1996
-- Version then    2.09
-- Support then    Use the mailing list twsocket@rtfm.be See website for details.
-- Legal issues then Copyright (C) 1996-2002 by Franois PIETTE
--         Rue de Grady 24, 4053 Embourg, Belgium. Fax then +32-4-365.74.56
--         <francois.piette@pophost.eunet.be>
--
--         This software is provided "as-is", without any express or
--         implied warranty.  In no event will the author be held liable
--         for any  damages arising from the use of this software.
--
--         Permission is granted to anyone to use this software for any
--         purpose, including commercial applications, and to alter it
--         and redistribute it freely, subject to the following
--         restrictions then
--
--         1. The origin of this software must not be misrepresented,
--          you must not claim that you wrote the original software.
--          If you use this software in a product, an acknowledgment
--          in the product documentation would be appreciated but is
--          not required.
--
--         2. Altered source versions must be plainly marked as such, and
--          must not be misrepresented as being the original software.
--
--         3. This notice may not be removed or altered from any source
--          distribution.
--
--         4. You must register this software by sending a picture postcard
--          to the author. Use a nice stamp and mention your name, street
--          address, EMail address and any comment you like to say.
--
-- Updates then
-- Jul 22, 1997 Adapted to Delphi 3
-- Sep 5, 1997  Added version information, removed old code, added OnTermType
--        Renamed some indentifier to be more standard.
-- Sep 24, 1997 V2.03 Added procedures to negociate options
-- May 12, 1998 V2.04 Changed NegociateOption to properly handle unwanted
--        option as Jan Tomasek <xtomasej@feld.cvut.cz> suggested.
-- Aug 10, 1998 V2.05 Cleared tcpSubOption after NegociateSubOption as Jan
--        Tomasek <xtomasej@feld.cvut.cz> suggested.
-- Aug 15, 1999 V2.06 Moved Notification procedure to global section for
--        BCB4 compatibility
-- Aug 20, 1999 V2.07 Added compile time options. Revised for BCB4.
-- Jun 18, 2001 V2.08 Use AllocateHWnd and DeallocateHWnd from wsocket.
-- Oct 23, 2002 V2.09 Changed Buffer arg in OnDataAvailable to Pointer instead
--        of PChar to avoid Delphi 7 messing everything with AnsiChar.
-- Oct 23, 2015 V2.09 Ported to OpenEuphoria 4 by Jean-Marc DURO
-- Mar 31, 2016 V2.09 Ported to Euphoria 3.11 by Jean-Marc DURO

include dll.e
include Sockets.ew
include WSErrors.ew
include Win32Lib.ew  -- doEvents
include myDebug.e
include myConv.e
include mySeq.e as seq
include myFile.e
include common.e
include buffer.e
include file.e

constant
--  TnCnxVersion     = 209,
--  CopyRight = " TTnCnx (c) 1996-2002 F. Piette V2.09 ",
  FLocation = "TNCNX",
  FTermType = "VT100",

  -- Telnet command characters                       
  TNCH_EOR        = 239,  -- #EF End Of Record (preceded by tcpIAC)     
  TNCH_SE         = 240,  -- #F0 End of subnegociation parameters    
--  TNCH_NOP        = 241,  -- #F1 No operation
--  TNCH_DATA_MARK  = 242,  -- #F2 Data stream portion of a Synch
--  TNCH_BREAK      = 243,  -- #F3 NVT charcater break
  TNCH_IP         = 244,  -- #F4 Interrupt process           
  TNCH_AO         = 245,  -- #F5 Abort output              
  TNCH_AYT        = 246,  -- #F6 Are you there             
  TNCH_EC         = 247,  -- #F7 Erase character             
  TNCH_EL         = 248,  -- #F8 Erase line              
--  TNCH_GA         = 249,  -- #F9 Go ahead
  TNCH_SB         = 250,  -- #FA Subnegociation            
  TNCH_WILL       = 251,  -- #FB Will                  
  TNCH_WONT       = 252,  -- #FC Wont                  
  TNCH_DO         = 253,  -- #FD Do                  
  TNCH_DONT       = 254,  -- #FE Dont                  
  TNCH_IAC        = 255,  -- #FF tcpIAC                   

  -- Telnet options                            
  TN_TRANSMIT_BINARY = 0,   -- #00 
  TN_ECHO            = 1,   -- #01 
--  TN_RECONNECTION    = 2,   -- #02
  TN_SUPPRESS_GA     = 3,   -- #03 
--  TN_MSG_SZ_NEGOC    = 4,   -- #04
--  TN_STATUS          = 5,   -- #05
--  TN_TIMING_MARK     = 6,   -- #06
--  TN_NOPTIONS        = 6,   -- #06
--  TN_DET             = 20,  -- #14
  TN_SEND_LOC        = 23,  -- #17 
  TN_TERMTYPE        = 24,  -- #18 
  TN_EOR             = 25,  -- #19 
--  TN_NAWS            = 31,  -- #1F
--  TN_TERMSPEED       = 32,  -- #20
--  TN_TFC             = 33,  -- #21
--  TN_XDISPLOC        = 35,  -- #23
--  TN_EXOPL           = 255, -- #FF

  TN_TTYPE_SEND      = 1,
  TN_TTYPE_IS        = 0


global constant DISCONNECTED = 0, CONNECTED = 1

global atom tcpSocket, tcpLastIO
  tcpSocket = 0
  tcpLastIO = time()

global sequence tcpHost, tcpSubOption
  tcpHost = ""
  tcpSubOption = ""

global integer tcpPort, tcpStatus, tcpBuffer
  tcpPort = 0
  tcpStatus = DISCONNECTED
  tcpBuffer = 0

integer tcpIAC, tcpVerb, tcpSubNegoc
  tcpIAC = 0
  tcpVerb = #0
  tcpSubNegoc = 0

global integer onTheFlowDisplayRoutine
onTheFlowDisplayRoutine = 0

constant MAXLEN = 1000

constant maxTime = 30

integer
  RemoteBinMode,
  LocalBinMode,
  FLocalEcho,
  spga,
  FTType

integer displayBuffer
displayBuffer = bufferNew()

------------------------------------------------------------------------------

global procedure handleError()
  integer err
  sequence msg, title, id, short

  err   = WSAGetLastError()
  id    = WSGetErrorString(err)
  short = WSGetShortDescription(err)
  title = sprintf("Winsock Error %d", err)
  msg   = sprintf("%s: %s", {id, short})
  VOID  = message_box(msg, title, MB_OK)
end procedure

------------------------------------------------------------------------------

global function getPort(atom hsock)
  sequence Address

  -- Get the port that is bound to the socket
  Address = getsockname (hsock)
  return Address[4] + Address[3] * 256
end function

------------------------------------------------------------------------------

global procedure sockDisconnect ()
-- Close the connection if one already exists.
  -- Disconnect the existing socket
  VOID = closesocket (tcpSocket)
end procedure

------------------------------------------------------------------------------

global function sockConnect (sequence Host, integer Port)
  object Address
  atom hsock

  -- Create a stream (TCP) socket
  hsock = socket (AF_INET, SOCK_STREAM, 0)
  if hsock != SOCKET_ERROR then
    -- Setup an IP address/port structure
    -- Use the remote address and port specified in the edit boxes
    Address = sockaddr_in (AF_INET, Host, Port)
    if sequence(Address) then
      -- Connect the new socket to the remote host
      if connect (hsock, Address) != SOCKET_ERROR then
        return hsock
      end if
    end if
  end if

  handleError()
  VOID = closesocket (hsock)
  return 0
end function

--------------------------------------------------------------------------------

global function telnetSend(sequence cmd)
  integer  status
  
  logMsg(sprintf("telnetSend(%s)", {cmd}))
  if not tcpSocket then return 0 end if
  status = send(tcpSocket, cmd)
  if status != SOCKET_ERROR then
    logMsg(sprintf("telnetSend: Sent '%s' %s (%d bytes)", {showPrintable(cmd), sprint(cmd), status}))
    tcpLastIO = time()
  else
    handleError()
    logMsg(sprintf("telnetSend: Error while sending '%s'", {cmd}))
  end if
  return status
end function

--------------------------------------------------------------------------------

procedure Answer(integer chAns, integer chOption)
  sequence Buf

  Buf = {TNCH_IAC, chAns, chOption}
  void = telnetSend(Buf)
end procedure

--------------------------------------------------------------------------------

procedure NegociateSubOption(sequence strSubOption)
  sequence Buf

  if strSubOption[1] = TN_TERMTYPE then
    if strSubOption[2] = TN_TTYPE_SEND then
      Buf = {TNCH_IAC, TNCH_SB, TN_TERMTYPE, TN_TTYPE_IS} & FTermType & {TNCH_IAC, TNCH_SE}
      void = telnetSend(Buf)
    end if
  else
--    puts(f_debug, "Unknown suboption\r\n"); 
  end if
end procedure

--------------------------------------------------------------------------------

procedure NegociateOption(integer chAction, integer chOption)
  sequence Buf
  
  if chOption = TN_TRANSMIT_BINARY then
    if chAction = TNCH_WILL then
      Answer(TNCH_DO, chOption)
      RemoteBinMode = 1
      LocalBinMode  = 1
    elsif chAction = TNCH_WONT then
      if RemoteBinMode then
        RemoteBinMode = 0
        LocalBinMode  = 0
      end if
    end if
  elsif chOption = TN_ECHO then
    if chAction = TNCH_WILL then
      Answer(TNCH_DO, chOption)
      FLocalEcho = 0
    elsif chAction = TNCH_WONT then
      FLocalEcho = 1
    end if
  elsif chOption = TN_SUPPRESS_GA then
    if chAction = TNCH_WILL then
      Answer(TNCH_DO, chOption)
      spga = 1
    end if
  elsif chOption = TN_TERMTYPE then
    if chAction = TNCH_DO then
      Answer(TNCH_WILL, chOption)
      FTType = 1
    end if
  elsif chOption = TN_SEND_LOC then
    if chAction = TNCH_DO then
      Answer(TNCH_WILL, chOption)
      Buf = {TNCH_IAC, TNCH_SB, TN_SEND_LOC, FLocation, TNCH_IAC, TNCH_SE}
      void = telnetSend(Buf)
    end if
  elsif chOption = TN_EOR then
    if chAction = TNCH_DO then
      Answer(TNCH_WILL, chOption)
      FTType = 1
    end if
  else
--    Answer(TNCH_WONT, chOption);
    -- Jan Tomasek <xtomasej@feld.cvut.cz>
    if chAction = TNCH_WILL then
      Answer(TNCH_DONT, chOption)
    else
      Answer(TNCH_WONT, chOption)
    end if
  end if
end procedure

--------------------------------------------------------------------------------

procedure AddChar(integer Ch)
  bufferWrite(displayBuffer, {Ch})
  bufferWrite(tcpBuffer, {Ch})
end procedure

--------------------------------------------------------------------------------

procedure ReceiveChar(integer Ch)

  if tcpVerb != #0 then
    NegociateOption(tcpVerb, Ch)
    tcpVerb     = #0
    tcpSubOption = ""
    return
  end if

  if tcpSubNegoc then
    if Ch = TNCH_SE then
      tcpSubNegoc  = 0
      NegociateSubOption(tcpSubOption)
      tcpSubOption = ""
    else
      tcpSubOption = tcpSubOption & Ch
    end if
    return
  end if

  if tcpIAC then
    if Ch = TNCH_IAC then
      AddChar(Ch)
      tcpIAC = 0
    elsif find(Ch, {TNCH_DO, TNCH_WILL, TNCH_DONT, TNCH_WONT}) then
      tcpIAC   = 0
      tcpVerb = Ch
    elsif Ch = TNCH_EOR then
      puts(f_debug, "TNCH_EOR\r\n")
      tcpIAC   = 0
    elsif Ch = TNCH_SB then
--      puts(f_debug, "Subnegociation\r\n");
      tcpSubNegoc = 1
      tcpIAC    = 0
    else
      printf(f_debug, "Unknown %02x\r\n", Ch)
      tcpIAC = 0
    end if
    return
  end if

  if Ch = TNCH_EL then
    puts(f_debug, "TNCH_EL\r\n")
    AddChar(Ch)
  elsif Ch = TNCH_EC then
    puts(f_debug, "TNCH_EC\r\n")
    AddChar(Ch)
  elsif Ch = TNCH_AYT then
    puts(f_debug, "TNCH_AYT\r\n")
    AddChar(Ch)
  elsif Ch = TNCH_IP then
    puts(f_debug, "TNCH_IP\r\n")
    AddChar(Ch)
  elsif Ch = TNCH_AO then
    puts(f_debug, "TNCH_AO\r\n")
    AddChar(Ch)
  elsif Ch = TNCH_IAC then
    tcpIAC = 1
  else
    AddChar(Ch)
  end if
end procedure

--------------------------------------------------------------------------------

global procedure telnetClose()
  if tcpStatus = DISCONNECTED then
    logMsg("telnetClose: already disconnected")
    return
  end if
  void = closesocket (tcpSocket)
  tcpStatus = DISCONNECTED
  tcpSocket = 0
end procedure

--------------------------------------------------------------------------------

procedure telnetReceive()
  object data

  logMsg("telnetReceive()")
  if not tcpSocket then return end if
  data = recv(tcpSocket, MAXLEN)
  if data[1] = 0 then -- indicates the connection was closed
    logMsg("telnetReceive: lost connection")
    tcpStatus = DISCONNECTED
    telnetClose()
  elsif data[1] > 0 then -- number of bytes received
--    printf(f_debug, "bytes received = %d\n", {data[1]})
    analyzeObject(data[2], "telnetReceive()", f_debug, 0)
    tcpLastIO = time()
    for i = 1 to length(data[2]) do ReceiveChar(data[2][i]) end for
    data = bufferRead(displayBuffer)
    if sequence (data) and onTheFlowDisplayRoutine then  -- process data with routine onTheFlowDisplayRoutine
      call_proc(onTheFlowDisplayRoutine, {data})
    end if
    bufferClear(displayBuffer)
  else
    handleError()
    logMsg("telnetReceive: Error while receiving")
    telnetClose()
  end if
end procedure

--------------------------------------------------------------------------------

-- has to be launched periodically by a 0.1 s timer
global procedure checkConnection(integer timeout, integer timeout_micro)
  -- get socket state
  object sock_data

--    logMsg(sprintf("tcpStatus = %d\n", {tcpStatus}))
--    printf(f_debug, "elapsed time = %d\n", {time() - tcpLastIO})
  if tcpStatus = CONNECTED then
    if (time() - tcpLastIO) >= maxTime then
      void = telnetSend({32,8})  -- SPACE and BACKSPACE
    end if
  end if
  sock_data = select({tcpSocket}, {}, {tcpSocket}, {timeout, timeout_micro})
--  analyzeObject(sock_data, "sock_data", f_debug, 0)
  if sock_data[1] = 0 then
    return -- no data received
  end if
  if length(sock_data[2]) then
    --logMsg("checkConnections: Data readable")
    telnetReceive()
  end if
  if length(sock_data[4]) then
    logMsg("checkConnections: Connexion is in error state, closed")
    telnetClose()
  end if
end procedure

--------------------------------------------------------------------------------

global function telnetConnect(sequence host, integer port)
  atom hsock
  
  hsock = sockConnect(host, port)
  if hsock = 0 then
    logMsg(sprintf("telnetConnect: Could not connect to %s:%d", {host, port}))
  else
    logMsg(sprintf("telnetConnect: Connected to %s:%d", {host, port}))
    tcpSocket = hsock
    tcpHost = host
    tcpPort = port
    tcpStatus = CONNECTED
    tcpLastIO = time()
    tcpBuffer = bufferNew()
    tcpIAC = 0
    tcpVerb = #0
    tcpSubOption = ""
    tcpSubNegoc = 0
  end if
  return hsock
end function

------------------------------------------------------------------------------

global function telnetRead(atom waitTime)
  -- wait waitTime
  -- read buffer
  -- display
  sequence data

--  logMsg(sprintf("telnetRead(%.3f)", {waitTime}))
  if not tcpSocket then return 0 end if
  puts(1, ".")
  pause(waitTime)
  data = bufferRead(tcpBuffer)
  bufferClear(tcpBuffer)
  return data
end function

------------------------------------------------------------------------------

global function readUntil(sequence invite, sequence other, atom waitTime, integer maxTime, integer debugMode)
  -- on lit avec telnetRead aprs waitTime
  -- tant qu'on n'a pas reu ce qui est attendu, on attend encore waitTime et ainsi de suite jusqu' ce qu'on ait atteint maxTime
  -- ds qu'on a reu ce qui est attendu on quitte
  sequence read
  atom t0
  object data

  logMsg(sprintf("readUntil('%s', %s, %.3f, %d)", {invite, dumpSequence(other), waitTime, maxTime}))
  if length(invite)=0 then
    logMsg("readUntil: actual invite missing !")
    return 0
  end if
  t0 = time()
  read = ""
  while 1 do
    checkConnection(0, 0)
    data = telnetRead(waitTime)
--    analyzeObject(data, "readUntil(): data", f_debug, 0)
    if atom(data) then return 0 end if
    if length(data) then
      read &= data
      analyzeObject(read, "readUntil(): read", f_debug, 0)
      if match(invite, read) then
        return read
      elsif length(other) and match_list(other, read, 1) then
        logMsg(sprintf("readUntil: failed to read '%s'", {invite}))
        return read
      end if
      if debugMode then	-- en mode debug, pas d'invite sans retour chariot
        if time()-t0 > 20 then -- toutes les 20 secondes
          void = telnetSend("\n")
          t0 = time()
        end if
      end if
    end if
    if (maxTime != -1) and ( (time()-t0) > maxTime ) then
      logMsg("readUntil(): maxTime exceeded")
      exit
    end if
  end while
  if length(data) then
    analyzeObject(data, "readUntil()", f_debug, 0)
  end if
  logMsg(sprintf("readUntil: failed to read '%s'", {invite}))
  return 0
end function

------------------------------------------------------------------------------

global function readAll(atom waitTime, object maxTime)
  -- on lit avec telnetRead aprs waitTime
  -- si on n'a rien reu, on attend encore waitTime et ainsi de suite jusqu' ce
  --   qu'on ait atteint maxTime si c'est un entier ou maxTime[1] (maxWaitTime) si c'est une squence
  -- si on a reu quelque chose, on attend waitTime et ainsi de suite tant qu'on reoit
  --   jusqu' ce qu'on ait atteint maxTime si c'est un entier ou maxTime[2] (maxReadTime) si c'est une squence
  --   on arrte de suite si on ne reoit plus
  sequence read
  atom t0
  object data
  integer flagReceiving

  logMsg(sprintf("readAll(%.3f, %s)", {waitTime, sprint(maxTime)}))
  t0 = time()
  read = ""
  flagReceiving = 0
  while 1 do
    checkConnection(0, 0)
    data = telnetRead(waitTime)
    if atom(data) then return 0 end if
    if length(data) then
      read &= data
--      analyzeObject(read, sprintf("readAll(%d): read", {cnx}), f_debug, 0)
      if not flagReceiving and sequence(maxTime) then
        t0 = time()
      end if
      flagReceiving = 1
      if atom(maxTime) then
        if (maxTime != -1) and ( (time()-t0) > maxTime ) then
          logMsg("readAll(): maxTime exceeded")
          exit
        end if
      else
        if (time()-t0) > maxTime[2] then
          logMsg("readAll(): maxReadTime exceeded")
          exit
        end if
      end if
    else
      if not flagReceiving then
        if atom(maxTime) then
          if (maxTime != -1) and ( (time()-t0) > maxTime ) then
            logMsg("readAll(): maxTime exceeded")
            exit
          end if
        else
          if (time()-t0) > maxTime[1] then
            logMsg("readAll(): maxWaitTime exceeded")
            exit
          end if
        end if
      else  --  on a reu mais on ne reoit plus: on arrte de suite
        exit
      end if      
    end if
  end while
  return read
end function

------------------------------------------------------------------------------

global function sendReceive(sequence cmd, sequence actualInvite, sequence loginInvite, atom waitTime, integer maxTime)
  logMsg(sprintf("sendReceive('%s', '%s', '%s', %.3f, %d)",{showPrintable(cmd), actualInvite, loginInvite, waitTime, maxTime}))
  if telnetSend(cmd) < 0 then return 0 end if
  return readUntil(actualInvite, loginInvite, waitTime, maxTime, 1)
end function
